unit HVObjectToString;

interface

{$IFNDEF CLR}
uses
  Classes;
{$ENDIF}

function ObjectToString(Value: TObject): string;

{$IFNDEF CLR}
const
  WM_TOSTRING = $7fff;

type
  TWMToString = packed record
    Msg: Cardinal;
    ToString: string;
  end;
  TObjectEx = class(TObject)
  public
    function ToString: string; virtual;
  end;
  TPersistentEx = class(TPersistent)
  public
    function ToString: string; virtual;
  end;
  TObjectToStringProc = function (Instance: TObject): string;

procedure RegisterObjectToString(AClassRef: TClass; AObjectToStringProc: TObjectToStringProc);
procedure UnregisterObjectToString(AClassRef: TClass; AObjectToStringProc: TObjectToStringProc);
function FindObjectToStringProc(AClassRef: TClass): TObjectToStringProc;

function TObject_Perform(Self: TObject; Msg: Cardinal; WParam, LParam: Longint): Longint;
function TObject_GetTextLen(Self: TObject): Integer;
function TObject_GetTextBuf(Self: TObject; Buffer: PChar; BufSize: Integer): Integer;
function TObject_GetText(Self: TObject): string;
function TObject_ToString(Self: TObject): string;

{$IFDEF WIN32CLASSHELPERS}
type
  TObjectHelper = class helper for TObject
  public
    function ToString: String;
  end;
{$ENDIF}

{$ENDIF}

implementation

{$IFDEF CLR}
function ObjectToString(Value: TObject): string;
begin
  if Assigned(Value)
  then Result := Value.ToString
  else Result := 'nil';
end;
{$ELSE}
uses
  Messages, SysUtils, Contnrs;

function TObject_Perform(Self: TObject; Msg: Cardinal; WParam, LParam: Longint): Longint;
var
  Message: TMessage;
begin
  Message.Msg := Msg;
  Message.WParam := WParam;
  Message.LParam := LParam;
  Message.Result := 0;
  if Self <> nil then Self.Dispatch(Message);
  Result := Message.Result;
end;

function TObject_GetTextLen(Self: TObject): Integer;
begin
  Result := TObject_Perform(Self, WM_GETTEXTLENGTH, 0, 0);
end;

function TObject_GetTextBuf(Self: TObject; Buffer: PChar; BufSize: Integer): Integer;
begin
  Result := TObject_Perform(Self, WM_GETTEXT, BufSize, Longint(Buffer));
end;

function TObject_GetText(Self: TObject): string;
var
  Len: Integer;
begin
  Len := TObject_GetTextLen(Self);
  SetLength(Result, Len);
  if Len <> 0 then
    TObject_GetTextBuf(Self, Pointer(Result), Len + 1);
end;

function TObject_ToString(Self: TObject): string;
var
  Message: TWMToString;
begin
  Message.Msg := WM_TOSTRING;
  if Self <> nil then Self.Dispatch(Message);
  Result := Message.ToString;
end;

var
  ObjectToStringList: TObjectList;

type
  TObjStr = class
    ClassRef: TClass;
    ObjectToStringProc: TObjectToStringProc;
    constructor Create(AClassRef: TClass; AObjectToStringProc: TObjectToStringProc);
  end;

constructor TObjStr.Create(AClassRef: TClass; AObjectToStringProc: TObjectToStringProc);
begin
  inherited Create;
  ClassRef := AClassRef;
  ObjectToStringProc := AObjectToStringProc;
end;

procedure RegisterObjectToString(AClassRef: TClass; AObjectToStringProc: TObjectToStringProc);
begin
  ObjectToStringList.Add(TObjStr.Create(AClassRef, AObjectToStringProc));
end;

procedure UnregisterObjectToString(AClassRef: TClass; AObjectToStringProc: TObjectToStringProc);
var
  I: Integer;
  ObjStr: TObjStr;
begin
  for I := ObjectToStringList.Count-1 downto 0 do
  begin
    ObjStr := TObjStr(ObjectToStringList.List[I]);
    if (ObjStr.ClassRef = AClassRef) and (@ObjStr.ObjectToStringProc = @AObjectToStringProc) then
    begin
      ObjectToStringList.Delete(I);
      Exit;
    end;
  end;
end;

function FindObjectToStringProc(AClassRef: TClass): TObjectToStringProc;
var
  I: Integer;
  ObjStr: TObjStr;
begin
  for I := ObjectToStringList.Count-1 downto 0 do
  begin
    ObjStr := TObjStr(ObjectToStringList.List[I]);
    if AClassRef.InheritsFrom(ObjStr.ClassRef) then
    begin
      Result := ObjStr.ObjectToStringProc;
      Exit;
    end;
  end;
  Result := nil;
end;

function ObjectToString(Value: TObject): string;
var
  ValueClass: TClass;
  ObjectToStringProc: TObjectToStringProc;
begin
  if Assigned(Value) then
  begin
    ValueClass := Value.ClassType;
    ObjectToStringProc := FindObjectToStringProc(ValueClass);
    if Assigned(ObjectToStringProc) then
      Result := ObjectToStringProc(Value)
    else
    begin
      Result := TObject_GetText(Value);
      if Result = '' then
      begin
        Result := TObject_ToString(Value);
        if Result = '' then
          Result := Value.ClassName;
      end;
    end;
  end
  else
    Result := 'nil';
end;

{ TObjectHelper }

{$IFDEF WIN32CLASSHELPERS}
function TObjectHelper.ToString: String;
begin
  Result := ObjectToString(Self);
end;
{$ENDIF}

{ TPersistentEx }

function TPersistentEx.ToString: string;
begin
  Result := Format('%s (%x)', [ClassName, Integer(Self)]);
end;

{ TObjectEx }

function TObjectEx.ToString: string;
begin
  Result := Format('%s (%x)', [ClassName, Integer(Self)]);
end;

function TObjectExToString(Instance: TObject): string;
begin
  Result := TObjectEx(Instance).ToString;
end;

function TPersistentExToString(Instance: TObject): string;
begin
  Result := TPersistentEx(Instance).ToString;
end;

initialization
  ObjectToStringList := TObjectList.Create;
  RegisterObjectToString(TObjectEx, TObjectExToString);
  RegisterObjectToString(TPersistentEx, TPersistentExToString);

finalization
  FreeAndNil(ObjectToStringList);

{$ENDIF}
end.

