unit Queue;

interface

uses
  Borland.Vcl.Classes,
  System.Threading;

type
  TD4DNThreadSafeQueue = class
  private
    FDone        : boolean;
    FMaxSize     : integer;
    FDebug       : boolean;
    FQueue       : TStrings;
    FQueueMutex  : System.Threading.Mutex;
    FCondEmpty   : System.Object;
    FCondFull    : System.Object;

    procedure DebugMessage(const msg : string);
    procedure WaitForCondition( condition : System.Object );
    procedure TriggerCondition( condition : System.Object );
  public
    constructor Create(MaxSize : integer );
    destructor  Destroy; override;
    procedure   AddObject(obj : TObject);
    function    GetObject : TObject;
    property    Debug : boolean read FDebug write FDebug;
  end;


implementation

uses
  Borland.Vcl.SysUtils;

{ TD4DNThreadSafeQueue }

procedure TD4DNThreadSafeQueue.AddObject(obj: TObject);
var
  bEmpty : boolean;

begin
  DebugMessage('AO: Aquiring queue mutex1');
  FQueueMutex.WaitOne;
  try
    while (FQueue.Count = FMaxsize) and (not FDone) do
    begin
      DebugMessage('AO: releasing queue mutex1');
      FQueueMutex.ReleaseMutex;
      DebugMessage('Waiting for full condition to be triggered');
      WaitForCondition( FCondFull );
      DebugMessage('AO: aquiring queue mutex2');
      FQueueMutex.WaitOne;
    end;

    if not FDone then
    begin
      // see if we will need to signal the empty condition later
      bEmpty := FQueue.Count = 0;
      // add the object to the list
      FQueue.InsertObject(0,'',obj);

      DebugMessage('AO: releasing queue mutex2');
      FQueueMutex.ReleaseMutex;

      // now signal the empty condition if necessary
      if bEmpty then
      begin
        DebugMessage('Signalling the empty condition');
        TriggerCondition(FCondEmpty);
      end;

      DebugMessage('AO: acquiring queue mutex3');
      FQueueMutex.WaitOne;
    end;
  finally
    DebugMessage('AO: releasing queue mutex3');
    FQueueMutex.ReleaseMutex;
  end;
end;


constructor TD4DNThreadSafeQueue.Create(MaxSize : integer);
begin
  inherited Create;

  FDone    := false;
  FMaxSize := MaxSize;

  // create the list
  FQueue          := TStringList.Create;
  FQueue.Capacity := FMaxSize;

  // create the lock for the data of the queue
  FQueueMutex  := System.Threading.Mutex.Create;

  // create the locks for the monitors
  FCondFull      := System.Object.Create;
  FCondEmpty     := System.Object.Create;
end;

destructor TD4DNThreadSafeQueue.Destroy;
begin
  DebugMessage('Destructor called');
  FDone := true;

  // notify any outstanding waiters..
  DebugMessage('Signalling both monitors (destructor)');
  TriggerCondition(FCondEmpty);
  TriggerCondition(FCondFull);

  FCondEmpty.Free;
  FCondFull.Free;
  FQueueMutex.Free;
  FQueue.Free;

  inherited;
end;

function TD4DNThreadSafeQueue.GetObject: TObject;
var
  bFull : boolean;
  slot  : integer;

begin
  Result := nil;
  DebugMessage('GetObject: aquiring queue mutex1');
  FQueueMutex.WaitOne;
  try
    while (FQueue.Count = 0) and (not FDone) do
    begin
      DebugMessage('GO: releasing queue mutex1');
      FQueueMutex.ReleaseMutex;
      DebugMessage('GetObject: Waiting for the empty condition to be signalled');
      WaitForCondition(FCondEmpty);
      DebugMessage('GO: aquiring queue mutex2');
      FQueueMutex.WaitOne;
    end;

    if not FDone then
    begin
      bFull := FQueue.Count = FMaxSize;

      slot := Fqueue.Count-1;
      Result := FQueue.Objects[slot];
      FQueue.Delete(slot);

      if bFull then
      begin
        DebugMessage('GO: releasing queue mutex2');
        FQueueMutex.ReleaseMutex;
        DebugMessage('GetObject: signalling full condition');
        TriggerCondition(FCondFull);

        DebugMessage('GO: acquiring queue mutex3');
        FQueueMutex.WaitOne;
      end;
    end;
  finally
    DebugMessage('GO: releasing queue mutex3');
    FQueueMutex.ReleaseMutex;
  end;
end;

procedure TD4DNThreadSafeQueue.WaitForCondition( condition : System.Object );
begin
  DebugMessage('WFC: Waiting for condition');
  System.Threading.Monitor.Enter(condition);
  try
    System.Threading.Monitor.Wait(condition);
  finally
    System.Threading.Monitor.Exit(condition);
  end;
  DebugMessage('WFC: Wait is over');
end;

procedure TD4DNThreadSafeQueue.TriggerCondition( condition : System.Object );
begin
  DebugMessage('TC: Triggering condition');
  System.Threading.Monitor.Enter(condition);
  try
    System.Threading.Monitor.Pulse(condition);
  finally
    System.Threading.Monitor.Exit(condition);
  end;
  DebugMessage('TC: Trigger is over');
end;

procedure TD4DNThreadSafeQueue.DebugMessage(const msg : string);
begin
  if FDebug then
    writeln(AppDomain.GetCurrentThreadID(),' ',msg);
end;

end.
